#!/usr/bin/env python

import math
from math import sqrt, sin, cos

import pyglet
from pyglet.gl import *
from pyglet.window import key

import cocos
from cocos import actions
from cocos.director import director

from gamelib.const import *
from gamelib import util
from gamelib import guiutil
from gamelib.res import *
from gamelib.keys import keys


class MoveDolphin(actions.WrappedMove):
    def __init__(self, *args, **kwargs):
        super(MoveDolphin, self).__init__(*args, **kwargs)
        self.max_speed = 500
        self.max_accel = 500

        # True if the dolphin has regained control (i.e. accelerated) after returning to water
        self.controlled = True

    def step(self, dt):
        dolphin = self.target
        vx, vy = dolphin.velocity
        r = math.radians(dolphin.rotation)
        cos_r, sin_r = cos(r), sin(r)

        # Wrap only at sides
        if dolphin.y - dolphin.width / 2 < 0:
            dolphin.y = dolphin.width / 2
            dolphin.velocity = 0, 0
        elif dolphin.y + dolphin.width / 2 > WINDOW_HEIGHT:
            dolphin.y = WINDOW_HEIGHT - dolphin.width / 2
            dolphin.velocity = 0, 0

        dolphin.dr = (keys[key.RIGHT] - keys[key.LEFT]) * 175

        # Dolphin in water is more transparent and with less gravity
        if dolphin.y > WATER_MARK:
            if dolphin.opacity != 255:
                dolphin.opacity = 255
                guiutil.splash(dolphin, entering=False)
            dolphin.acceleration = util.sign(vx) * -100, -400 if dolphin.y + dolphin.width < WINDOW_HEIGHT -50 else -700
            self.controlled = False
        else:
            if dolphin.opacity != UNDERWATER_OPACITY:
                splash_sound.play()
                # Dolphin has just returned to water
                dolphin.opacity = UNDERWATER_OPACITY
                guiutil.splash(dolphin)

                # Rotate then return control a little after returning to water
                rotation = dolphin.rotation % 360
                if rotation < 45:
                    dr = -rotation / 2
                elif rotation < 135:
                    dr = (90 - rotation) / -2
                elif rotation < 215:
                    dr = (180 - rotation) / 2
                else:
                    dr = (270 - rotation) / -2

                def control():
                    self.controlled = True
                dolphin.do(actions.RotateBy(dr, 0.4) + actions.CallFunc(control))


            # SPACE to accelerate but only in water.
            if self.controlled and keys[key.SPACE]:
                # Acceleration direction is direction dolphin is facing
                dolphin.acceleration = -cos_r * self.max_accel, sin_r * self.max_accel
            else:
                dolphin.acceleration = -vx * 3, -vy * 3
            
            # Slight drop in water
            dolphin.acceleration = dolphin.acceleration[0], dolphin.acceleration[1] - 250

            # Limit speed and ensure it's in direction dolphin is facing if controlled
            if self.controlled:
                speed = min(sqrt(vx ** 2 + vy ** 2), self.max_speed)
                vx, vy = -cos_r * speed, sin_r * speed

        dolphin.velocity = vx, vy

        # Show rings the dolphin is carrying (max 6)
        if dolphin.ring_count >= director.scene.required_rings:
            if dolphin.ring_count == director.scene.required_rings:
                for ring in dolphin.rings:
                    ring.kill()
                dolphin.rings = []
            dolphin.all_rings.x, dolphin.all_rings.y = dolphin.x, dolphin.y
            dolphin.all_rings.opacity = dolphin.opacity
            dolphin.all_rings.rotation = dolphin.rotation
        else:
            shown_rings = min(MAX_RINGS, dolphin.ring_count)
            for ring in dolphin.rings[:shown_rings]:
                ring.x, ring.y = dolphin.x, dolphin.y
                ring.opacity = dolphin.opacity
                ring.rotation = dolphin.rotation

#        director.scene.debug_label.element.text = 'v=%s, r=%s' % (str(dolphin.velocity), dolphin.rotation)

        super(MoveDolphin, self).step(dt)

class Dolphin(cocos.sprite.Sprite):

    def __init__(self, image, rings, all_rings):
        super(Dolphin, self).__init__(image)

        # Cocos2d attributes
        self.position = 500, 200
        self.velocity = (-30, 0)
        self.acceleration = (0, 0)
        self.rotation = 0
        self.opacity = UNDERWATER_OPACITY

        # The rings the dolphin is carrying
        self.ring_count = 0
        self.rings = rings
        self.all_rings = all_rings

        self.do(MoveDolphin(WINDOW_WIDTH, WINDOW_HEIGHT))

    def add_ring(self):
        """Add another ring for the dolphin to carry."""
        # No need to add a real ring as we're only going to show the first few
        self.ring_count += 1

    def caught(self, ring):
        """Return True if ring has been caught.
        
        The dolphin's nose must pass through the ring.
        """
        return util.rect_hit( self.nose_pos(),
                             (ring.x, ring.y),
                             self.rotation,
                             ring.width, 
                             ring.height)

    def nose_pos(self):
        """The position of the dolphin's nose."""
        angle = math.radians(self.rotation)
        x = self.x - self.width / 2 * cos(angle) - NOSE_OFFSET * sin(angle)
        y = self.y + self.width / 2 * sin(angle) - NOSE_OFFSET * cos(angle)
        return x, y

    def draw(self):
        super(Dolphin, self).draw()

        # Draw a red dot on nose.  Originally for debugging, but I like it!
        x, y = self.nose_pos()
        glLineWidth(2)
        glBegin(GL_LINE_LOOP);
        glColor3f(1, 0, 0)
        glVertex3f(x, y, 0);
        glVertex3f(x + 1, y + 1, 0);
        glEnd();
        glFlush();
