#!/usr/bin/env python

import sys
import math
from math import sin, cos
import random
import itertools

import pyglet
from pyglet.gl import *
from pyglet.window import key
pyglet.resource.path = ['@data']
pyglet.resource.reindex()

import cocos
from cocos import actions
from cocos.director import director
from cocos.scenes import FadeTransition

from gamelib.util import *
from gamelib import guiutil
from gamelib.const import *
from gamelib.res import *
from gamelib.dolphin import Dolphin
from gamelib.keys import keys
from gamelib.menu import MenuScene


class MoveRing(actions.Move):

    # The associated ring is on its way out
    dying = False

    def step(self, dt):
        super(MoveRing, self).step(dt)

        if self.dying:
            return

        ring = self.target
        dolphin = director.scene.dolphin

        # Ring in water is more transparent and falls slower at an angle
        if ring.y < WATER_MARK:
            if ring.opacity != UNDERWATER_OPACITY:
                # Just entered the water
                ring.opacity = UNDERWATER_OPACITY
                ring.gravity = 0
                ring.velocity = ring.velocity[0] / 2, ring.velocity[1] / 4
                guiutil.splash(ring, weight=1.5)
                ring.do(actions.FadeOut(0.5) + actions.CallFuncS(lambda s: s.kill))
                self.dying = True
                return

        if not 0 < ring.x < WINDOW_WIDTH or ring.y > WINDOW_HEIGHT:
            ring.kill()
        elif dolphin.caught(ring):
            # We've caught all the rings already
            if dolphin.ring_count == director.scene.required_rings:
                return

            dolphin.add_ring()
            ring_sound.play()

            # How high the ring was caught determines points
            if ring.y > 480:
                points = 30
            elif ring.y > 375:
                points = 10
            else:
                points = 0

            if points:
                director.scene.increment_score(points)
                points_label = cocos.text.Label('%s Points' % points, font_name=GAME_FONT, font_size=10, color=(0, 0, 0, 255), x=ring.x, y=ring.y)
                director.scene.hud_layer.add(points_label, z=200)
                points_label.do(actions.MoveBy((20, 10), duration=0.7) | (actions.Delay(0.2) + actions.FadeOut(.5) + actions.CallFuncS(lambda s: s.kill())))

            # Tick a sailor
            tick = cocos.sprite.Sprite(tick_image, position=((dolphin.ring_count - 1) * (voice_image.width + 10) + 30, 40))
            tick.opacity = 0
            director.scene.ring_layer.add(tick, z=200)
            tick.do(actions.FadeIn(0.5))

            # We've now caught enough rings to get to the next level
            if dolphin.ring_count == director.scene.required_rings:
                applause_sound.play()
                finish_text = 'You\'ve caught %s liferings and saved the sailors!' % dolphin.ring_count
                if ring.x < 200:
                    anchor_x = 'left'
                elif ring.x < 600:
                    anchor_x = 'center'
                else:
                    anchor_x = 'right'
                finish_label = cocos.text.Label(finish_text, font_name=GAME_FONT, font_size=13, color=(0, 0, 0, 255), position=ring.position, anchor_x=anchor_x, anchor_y='bottom')
                director.scene.hud_layer.add(finish_label, z=200)
                def new_level():
                    director.replace(LevelScene(level=director.scene.level + 1, score=director.scene.score))

                finish_label.do(actions.MoveBy((-30, 15), duration=4) | (actions.Delay(1.5) + actions.FadeOut(2.5) + actions.CallFuncS(lambda s: s.kill()) + actions.CallFunc(new_level)))

            ring.kill()

class MainWindow(cocos.scene.Scene):

    def __init__(self, level=1, score=0):
        super(MainWindow, self).__init__()

        self.level = level
        self.score = score
        self.required_rings = guiutil.required_rings(level)

        self.main_layer = cocos.layer.Layer()
        self.dolphin_layer = cocos.layer.Layer()
        self.ring_layer = cocos.layer.Layer()
        self.splash_layer = cocos.layer.Layer()
        self.hud_layer = cocos.layer.Layer()

        self.setup_background()
        self.boat = self.create_boat()

        for i in range(self.required_rings):
            self.ring_layer.add(cocos.sprite.Sprite(voice_image, (i * (voice_image.width + 10) + 30, 40)))

        # Dolphin
        rings = []
        for i in range(MAX_RINGS):
            ring = cocos.sprite.Sprite(caught_ring_image_grid[i])
            ring.opacity = 0
            rings.append(ring)
            self.ring_layer.add(ring)
        all_rings = cocos.sprite.Sprite(all_rings_image)
        all_rings.opacity = 0
        self.ring_layer.add(all_rings)
        self.dolphin = Dolphin(dolphin_image, rings, all_rings)
        self.dolphin_layer.add(self.dolphin)

        self.debug_label = cocos.text.Label('', font_size=10, x=20, y=20)
        self.score_label = cocos.text.Label('Score: %s' % score, font_name=GAME_FONT, font_size=20, color=(100, 100, 100, 255), x=20, y=WINDOW_HEIGHT - 20, anchor_x='left', anchor_y='top')
        self.level_label = cocos.text.Label('LEVEL %s' % level, font_name=GAME_FONT, font_size=20, color=(100, 100, 100, 255), x=WINDOW_WIDTH / 2, y=WINDOW_HEIGHT - 20, anchor_x='center', anchor_y='top')
        rings_required_label = cocos.text.Label('Catch %s liferings' % self.required_rings, font_name=GAME_FONT, font_size=12, color=(255, 255, 255, 255), x=30 - voice_image.width / 2, y=5, anchor_x='left', anchor_y='bottom')

        self.hint_label = cocos.text.Label('', font_name=GAME_FONT, font_size=15, x=WINDOW_WIDTH - 10, y=5, anchor_x='right', anchor_y='bottom')
        self.hint_label.opacity = 0
        self.hud_layer.add(self.debug_label)
        self.hud_layer.add(self.score_label)
        self.hud_layer.add(self.level_label)
        self.hud_layer.add(rings_required_label)
        self.hud_layer.add(self.hint_label)

        self.add(self.main_layer, z=10)
        self.add(self.splash_layer, z=20)
        self.add(self.dolphin_layer, z=30)
        self.add(self.ring_layer, z=40)
        self.add(self.hud_layer, z=50)

        self.start_clouds()
        self.throw_ring()
        if level == 1:
            self.start_hints()
        
    def setup_background(self):
        back = cocos.sprite.Sprite(background_image, anchor=(0, 0))
        sun = cocos.sprite.Sprite(sun_image, position=(720, 520))
        for z, i in enumerate([back, sun]):
            self.main_layer.add(i, z=z)

    def create_boat(self):
        boat = cocos.sprite.Sprite(boat_image)
        boat.position = 0, WATER_MARK + 75
        def boat_finished():
            pyglet.clock.unschedule(self.throw_ring)
            pyglet.clock.unschedule(self.start_clouds)
            # Game over?
            if self.dolphin.ring_count < self.required_rings:
                director.replace(GameOverScene(self.level, self.score))

        boat.do(actions.MoveTo((WINDOW_WIDTH + boat.width / 2, WATER_MARK + 75), 15 if guiutil.is_debug() else 50 - self.level * 5) + actions.CallFunc(boat_finished))
        self.main_layer.add(boat, z=10)
        return boat

    def start_clouds(self, dt=0):
        cloud = cocos.sprite.Sprite(cloud_image_grid[random.randrange(0, 3)])
        cloud.scale = random.random() * 0.5 + 0.5
        cloud.position = -cloud.width / 2, random.randrange(450, 600)
        v = random.randrange(10, 30)
        cloud.opacity = 50 + (v - 10) * 4
        self.main_layer.add(cloud, z=5)
        final_x = WINDOW_WIDTH + cloud.width
        cloud.do(actions.MoveTo((final_x, cloud.y), duration=(final_x - cloud.x) / v))
        pyglet.clock.schedule_once(self.start_clouds, random.random() * 8 + 4)

    def start_hints(self):
        def set_hint(text):
            def setter():
                self.hint_label.element.text = text
            return setter

        hint_actions = actions.Delay(1)
        for hint in HINTS:
            hint_actions += actions.CallFunc(set_hint(hint)) + actions.FadeIn(1) + actions.Delay(4) + actions.FadeOut(1) + actions.Delay(4) 
        self.hint_label.do(hint_actions)

    def throw_ring(self, dt=0):
        ring = cocos.sprite.Sprite(ring_image)
        ring.position = self.boat.position
        if ring.x < WINDOW_WIDTH / 3:
            direction = 1
        elif ring.x > WINDOW_WIDTH * 2 / 3:
            direction = -1
        else:
            direction = random.choice([-1, 1])
        
        difficulty = self.level * 10
        ring.velocity = direction * random.randint(60, 120 + int(difficulty / 1.5)), random.randint(175, 250)
        ring.gravity = -165 + difficulty
        ring.rotation = 90 + ring.velocity[0] / 2
        ring.opacity = 0
        self.ring_layer.add(ring)
        ring.do(MoveRing() | actions.FadeTo(255, 0.5))
        pyglet.clock.schedule_once(self.throw_ring, random.random() * 1.5 + 0.75)

    def increment_score(self, amount):
        self.score += amount
        self.score_label.element.text = 'Score: %s' % self.score

class LevelScene(MenuScene):
    def __init__(self, level=1, score=0):
        super(LevelScene, self).__init__('Level %s' % level)

        def enter_game():
            start_background_sound()
            director.replace(MainWindow(level=level, score=score))

        self.menu.create_menu([
            cocos.menu.MenuItem('Press ENTER to %s' % ('continue' if level > 1 else 'start'), enter_game),
        ])

        if level > 1:
            label_text = 'Well done!\nYou now need to catch %s liferings\nto save all the sailors at the bottom of the sea.' % guiutil.required_rings(level)
        else:
            label_text = 'You need to catch %s liferings\nto save all the sailors at the bottom of the sea.' % guiutil.required_rings(level)
        self.add_text(label_text)
        self.menu.on_quit = director.pop

class GameOverScene(MenuScene):
    def __init__(self, level, score):
        super(GameOverScene, self).__init__('Game Over!')

        self.menu.create_menu([
            cocos.menu.MenuItem('Press ENTER to return to menu', lambda: director.pop()),
        ])

        self.add_text('You reached LEVEL %s and scored %s points' % (level, score))
        self.menu.on_quit = director.pop

class StoryScene(MenuScene):
    def __init__(self):
        super(StoryScene, self).__init__('The Story')

        self.menu.create_menu([
            cocos.menu.MenuItem('Press ENTER to return to menu', lambda: director.pop()),
        ])

        self.add_text('You are Flapper the Friendly Dolphin.\n'
                      'Captain Bernard has lost his sailors at the bottom of the sea\n'
                      'and he is throwing them liferings to save them.\n'
                      'You have to help Bernard by catching the liferings so you\n'
                      'can take them to the sailors at the bottom of the sea.\n'
                      'You can save one sailor for every lifering you catch but do it\n'
                      'quickly ... you only have until Bernard\'s boat disappears!')

        self.add_text('Press SPACE to swim faster and use the ARROW keys to turn.\n'
                      'Try picking up speed by swimming horizontally then\n'
                      'turn to jump out of the water.  The faster you\n'
                      'swim, the higher you jump and the more points you get\n'
                      'when you catch a lifering.',
                      top=200)
        self.menu.on_quit = director.pop

class OpeningScene(MenuScene):

    def __init__(self):
        super(OpeningScene, self).__init__(GAME_TITLE)

        self.menu.create_menu([
            cocos.menu.MenuItem('Play', lambda: director.push(LevelScene())),
            cocos.menu.MenuItem('The Story', lambda: director.push(StoryScene())),
            cocos.menu.ToggleMenuItem('Music ', guiutil.volume, True),
            cocos.menu.MenuItem('Quit', pyglet.app.exit),
        ])
        labels = self.create_text('by Safe Hammad', top=40)
        for l in labels:
            l.opacity = 0
            self.add(l)
            l.do(actions.Delay(1) + actions.FadeIn(2))
        self.menu.on_quit = pyglet.app.exit

    def on_enter(self):
        super(OpeningScene, self).on_enter()
        stop_background_sound()

def main():
    #config = Config(sample_buffers=1, samples=4, depth_size=16, double_buffer=True)
    #director.init(WINDOW_WIDTH, WINDOW_HEIGHT, caption=GAME_TITLE, config=config)
    director.init(WINDOW_WIDTH, WINDOW_HEIGHT, caption=GAME_TITLE)
    director.window.push_handlers(keys)
    director.run(OpeningScene())
